import logging

from ..conf import settings


logger = logging.getLogger(__name__)


def post_python(conf, code, report, **kwargs):
    logger.info("Executing custom notifier")
    logger.debug(code)
    exec(code, {'text': report, 'conf': conf, 'creds': settings().creds})



import logging

import requests

from ..conf import settings


logger = logging.getLogger(__name__)


def post_gitter(conf, report, **kwargs):
    gitter = settings().notifiers.get('gitter', {})
    gitter.update(settings().creds.get('gitter', {}))
    response = requests.post(
        gitter['url'],
        data={"message": report},
    )
    logger.debug(response.text)
    response.raise_for_status()
    return response



import logging

import requests

from ..conf import settings


logger = logging.getLogger(__name__)


def post_mailgun(conf, report, **kwargs):
    mailgun = settings().notifiers.get('mailgun', {})
    mailgun.update(settings().creds.get('mailgun', {}))
    subject = "Kibitzer update for " + conf['name']
    response = requests.post(
        "https://api.mailgun.net/v3/{domain}/messages"
        .format(domain=mailgun['domain']),
        auth=("api", mailgun['key']),
        data={
            "from": "Kibitzer <mailgun@{domain}>"
                    .format(domain=mailgun['domain']),
            "to": [mailgun['to']],
            "subject": subject,
            "text": report,
        })
    logger.debug(response.text)
    response.raise_for_status()
    return response



from .mailgun import post_mailgun  # noqa
from .custom import post_python  # noqa
from .bash import post_bash  # noqa
from .gitter import post_gitter  # noqa
from .slack import SlackSession  # noqa
from .smtp import post_smtp  # noqa



from time import sleep
import requests

from kibitzr import __version__ as version


def simple_fetcher(conf):
    retries = 3
    for retry in range(retries):
        try:
            response = requests.get(
                url=conf['url'],
                headers={'User-agent': 'Kibitzer/' + version},
            )
        except requests.HTTPError:
            if retry == retries - 1:
                raise
            else:
                sleep(5)
                continue
        ok = (response.status_code == 200)
        return ok, response.text


class SessionFetcher(object):
    def __init__(self, conf):
        self.conf = conf
        self.session = requests.Session()
        self.session.headers.update({
            'User-agent': 'Kibitzer/' + version,
        })
        self.url = conf['url']
        self.valid_http = set(conf.get('valid_http', [200]))

    def fetch(self, *args, **_kwargs):
        retries = 3
        for retry in range(retries):
            try:
                response = self.session.get(self.url)
            except requests.HTTPError:
                if retry == retries - 1:
                    raise
                else:
                    sleep(5)
                    continue
            ok = (response.status_code in self.valid_http)
            return ok, response.text



import pytest
from .target.server import start_server, stop_server
from kibitzr.fetcher import cleanup_fetchers
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


server_addess = None


class SettingsMock(object):
    def __init__(self):
        self.pages = []
        self.notifiers = {}
        self.creds = {}


@pytest.fixture(scope="session", autouse=True)
def target_website(request):
    global server_addess
    server_process, server_addess = start_server()
    request.addfinalizer(cleanup_fetchers)
    request.addfinalizer(lambda: stop_server(server_process))
    patch_object = patch("kibitzr.fetcher.browser.settings",
                         return_value=SettingsMock())
    patch_object.start()
    request.addfinalizer(lambda: patch_object.stop())


@pytest.fixture
def target():
    global server_addess
    return server_addess


@pytest.fixture
def not_found_conf(target):
    return {
        'name': '404',
        'url': "http://{0}:{1}/page-not-found".format(*target),
    }


@pytest.fixture
def json_conf(target):
    return {
        'name': 'Test page',
        'url': "http://{0}:{1}/api.json".format(*target),
        'transform': ['json'],
    }


@pytest.fixture
def html_text_conf(target):
    return {
        'name': 'Index',
        'url': "http://{0}:{1}/index.html".format(*target),
        'transform': ['text'],
    }



import sh
import tempfile
import logging


logger = logging.getLogger(__name__)


def fetch_bash(conf, **kwargs):
    code = conf['script']
    logger.info("Executing bash fetcher")
    logger.debug(code)
    with tempfile.NamedTemporaryFile() as fp:
        logger.debug("Saving code to %r", fp.name)
        fp.write(code.encode('utf-8'))
        fp.flush()
        logger.debug("Launching script %r", fp.name)
        result = sh.bash(fp.name)
        stdout = result.stdout.decode('utf-8')
        stderr = result.stderr.decode('utf-8')
        logger.debug("Bash exit_code: %r", result.exit_code)
        logger.debug("Bash stdout: %s", stdout)
        logger.debug("Bash stderr: %s", stderr)
        ok = (result.exit_code == 0)
        if ok:
            report = stdout
        else:
            report = u'\n'.join([stdout, stderr])
        return ok, report



import functools
import logging
import traceback

from .fetcher import (
    firefox_fetcher,
    SessionFetcher,
    fetch_bash,
)
from .notifier import (
    SlackSession,
    post_bash,
    post_gitter,
    post_mailgun,
    post_python,
    post_smtp,
)
from .transformer import pipeline_factory


logger = logging.getLogger(__name__)


class Checker(object):
    def __init__(self, conf):
        self.conf = conf
        self.downloader = self.downloader_factory()
        self.transform_error = self.transform_error_factory()
        self.notifiers = self.create_notifiers()
        self.transform_pipeline = pipeline_factory(self.conf)

    @staticmethod
    def create_from_settings(checks, names=None):
        if names:
            selected_checks = [
                conf
                for conf in checks
                if conf['name'] in names
            ]
            selected_names = [conf['name'] for conf in selected_checks]
            if len(selected_checks) < len(checks):
                logger.info("Filtered list of checks to: %r",
                            ", ".join(sorted(selected_names)))
                checks = selected_checks
            if len(selected_names) < len(names):
                logger.error(
                    "Following check(s) were not found: %r",
                    ", ".join(sorted(set(names).difference(selected_names)))
                )
        return [Checker(conf)
                for conf in checks]

    def check(self):
        ok, content = self.fetch()
        ok, report = self.transform(ok, content)
        if not ok:
            report = self.transform_error(report)
        self.notify(report=report)
        return ok, report

    def fetch(self):
        if self.is_script():
            logger.info("Fetching %r using bash script",
                        self.conf['name'])
        else:
            logger.info("Fetching %r at %r",
                        self.conf['name'], self.conf['url'])
        try:
            ok, content = self.downloader(self.conf)
        except Exception:
            logger.exception(
                "Exception occured while fetching page"
            )
            ok, content = False, traceback.format_exc()
        return ok, content

    def downloader_factory(self):
        if self.needs_firefox():
            return firefox_fetcher
        elif self.is_script():
            return fetch_bash
        else:
            return SessionFetcher(self.conf).fetch

    def is_script(self):
        return all((
            'url' not in self.conf,
            'script' in self.conf,
        ))

    def needs_firefox(self):
        return any(
            self.conf.get(key)
            for key in ('scenario', 'delay')
        )

    def transform(self, ok, content):
        ok, content = self.transform_pipeline(ok, content)
        if content:
            content = content.strip()
        return ok, content

    def transform_error_factory(self):
        error_policy = self.conf.get('error', 'notify')
        if error_policy == 'ignore':
            return self.mute
        elif error_policy == 'notify':
            return self.echo
        else:
            logger.warning("Unknown error policy: %r", error_policy)
            logger.info("Defaulting to 'notify'")
            return self.echo

    @staticmethod
    def echo(content):
        # content will be logged in notifier
        logger.debug("Notifying on error")
        return content

    @staticmethod
    def mute(content):
        logger.debug("Ignoring error in %s", content)
        return None

    def create_notifiers(self):
        notifiers_conf = self.conf.get('notify', [])
        if not notifiers_conf:
            logger.warning(
                "No notifications configured for %r",
                self.conf['name'],
            )
        return list(filter(None, [
            self.notifier_factory(notifier_conf)
            for notifier_conf in notifiers_conf
        ]))

    @staticmethod
    def notifier_factory(notifier_conf):
        try:
            key, value = next(iter(notifier_conf.items()))
        except AttributeError:
            key, value = notifier_conf, None
        if key == 'python':
            return functools.partial(post_python, code=value)
        elif key == 'bash':
            return functools.partial(post_bash, code=value)
        elif key == 'mailgun':
            return post_mailgun
        elif key == 'gitter':
            return post_gitter
        elif key == 'slack':
            return SlackSession().post
        elif key == 'smtp':
            return functools.partial(post_smtp, notifier_conf=value)
        else:
            logger.error("Unknown notifier %r", key)

    def notify(self, report, **_kwargs):
        if report:
            logger.debug('Sending report: %r', report)
            for notifier in self.notifiers:
                try:
                    notifier(conf=self.conf, report=report)
                except Exception:
                    logger.exception(
                        "Exception occured during sending notification"
                    )
        else:
            logger.debug('Report is empty, skipping notification')



import sh
import tempfile
import logging


logger = logging.getLogger(__name__)


def post_bash(code, report, **kwargs):
    logger.info("Executing custom notifier")
    logger.debug(code)
    with tempfile.NamedTemporaryFile() as fp:
        logger.debug("Saving code to %r", fp.name)
        fp.write(code.encode('utf-8'))
        fp.flush()
        logger.debug("Launching script %r", fp.name)
        result = sh.bash(fp.name, _in=report)
        logger.debug("Bash exit_code: %r", result.exit_code)
        logger.debug("Bash stdout: %s", result.stdout.decode('utf-8'))
        logger.debug("Bash stderr: %s", result.stderr.decode('utf-8'))



import unittest
import tempfile
import shutil
from kibitzr.storage import PageHistory


class PageHistoryTestCase(unittest.TestCase):
    def setUp(self):
        self.storage_dir = tempfile.mkdtemp()
        self.history = PageHistory(
            conf={'name': 'test', 'url': 'web'},
            storage_dir=self.storage_dir,
        )

    def tearDown(self):
        shutil.rmtree(self.storage_dir)

    def test_unified_diff_sample(self):
        scenario = (
            (u"hello", True, u"test at web\n@@ -0,0 +1 @@\n+hello"),
            (u"world", True, u"test at web\n@@ -1 +1 @@\n-hello\n+world"),
            (u"world", False, None),
        )
        for content, changed, report in scenario:
            result = self.history.report_changes(content, verbose=False)
            assert result == (changed, report)

    def test_verbose_diff_sample(self):
        scenario = (
            (u"hello", True, u"test at web\nhello"),
            (u"world", True, u"test at web\nPrevious value:\nhello\nNew value:\nworld"),
            (u"world", False, None),
        )
        for content, changed, report in scenario:
            result = self.history.report_changes(content, verbose=True)
            assert result == (changed, report)



import os
import itertools
import io

import sh

from .utils import normalize_filename

git = sh.git.bake('--no-pager', _cwd="pages")


def report_changes(conf, content):
    return PageHistory(conf).report_changes(content)


class PageHistory(object):
    """
    Single file changes history using git
    """
    STORAGE_DIR = "pages"

    def __init__(self, conf, storage_dir=None):
        self.storage_dir = storage_dir or self.STORAGE_DIR
        self.cwd = os.path.join(
            self.storage_dir,
            normalize_filename(conf['name']),
        )
        self.target = os.path.join(self.cwd, "content")
        self.git = sh.git.bake(
            '--no-pager',
            _cwd=self.cwd,
        )
        self.ensure_repo_exists()
        self.commit_msg = "{name} at {url}".format(
            name=conf['name'],
            url=conf.get('url'),
        )

    def report_changes(self, content, verbose=False):
        """
        1) Write changes in file,
        2) Commit changes in git
        3.1) If something changed, return tuple(True, changes)
        3.2) If nothing changed, return tuple(False, None)
        If verbose is True, return changes in human-friendly format,
        else use unified diff
        """
        self.write(content)
        if self.commit():
            if verbose:
                return True, self.before_after()
            else:
                return True, self.last_log()
        else:
            return False, None

    def write(self, content):
        """Save content on disk"""
        with io.open(self.target, 'w', encoding='utf-8') as fp:
            fp.write(content)
            if not content.endswith(u'\n'):
                fp.write(u'\n')

    def commit(self):
        """git commit and return whether there were changes"""
        self.git('add', '-A', '.')
        try:
            self.git.commit('-m', self.commit_msg)
            return True
        except sh.ErrorReturnCode_1:
            return False

    def last_log(self):
        """Return last changes in truncated unified diff format"""
        output = self.git.log(
            '-1',
            '-p',
            '--no-color',
            '--format=%s',
        ).stdout.decode('utf-8')
        lines = output.splitlines()
        return u'\n'.join(
            itertools.chain(
                lines[:1],
                itertools.islice(
                    itertools.dropwhile(
                        lambda x: not x.startswith('+++'),
                        lines[1:],
                    ),
                    1,
                    None,
                ),
            )
        )

    def before_after(self):
        """Return changes in human-friendly format #14"""
        try:
            before = self.git.show('HEAD~1:content').strip()
        except sh.ErrorReturnCode_128:
            before = None
        after = self.git.show('HEAD:content').strip()
        if before is not None:
            return (u'{subject}\nPrevious value:\n{before}\nNew value:\n{after}'
                    .format(subject=self.commit_msg, before=before, after=after))
        else:
            return u'\n'.join([self.commit_msg, after])

    def ensure_repo_exists(self):
        """Create git repo if one does not exist yet"""
        if not os.path.isdir(self.cwd):
            os.makedirs(self.cwd)
        if not os.path.isdir(os.path.join(self.cwd, ".git")):
            self.git.init()
            self.git.config("user.email", "you@example.com")
            self.git.config("user.name", "Your Name")



import logging
import signal
import time
import code

import schedule

from .conf import settings
from .fetcher import cleanup_fetchers
from .checker import Checker


logger = logging.getLogger(__name__)
reload_conf_pending = False
interrupted = False
open_backdoor = False


def main(once=False, log_level=logging.INFO, names=None):
    global reload_conf_pending, interrupted
    logging.getLogger("").setLevel(log_level)
    logger.debug("Arguments: %r",
                 {"once": once, "log_level": log_level})
    signal.signal(signal.SIGUSR1, on_reload_config)
    signal.signal(signal.SIGUSR2, on_backdoor)
    signal.signal(signal.SIGINT, on_interrupt)
    try:
        while True:
            if interrupted:
                break
            if reload_conf_pending:
                settings().reread()
                reload_conf_pending = False
            checkers = Checker.create_from_settings(
                checks=settings().pages,
                names=names
            )
            execute_all(checkers)
            if once or interrupted:
                break
            else:
                check_forever(checkers)
    finally:
        cleanup_fetchers()


def check_forever(checkers):
    global reload_conf_pending, interrupted, open_backdoor
    schedule_checks(checkers)
    logger.info("Starting infinite loop")
    while not reload_conf_pending:
        if interrupted:
            break
        if open_backdoor:
            open_backdoor = False
            code.interact(
                banner="Kibitzr debug shell",
                local=locals(),
            )
        schedule.run_pending()
        if interrupted:
            break
        time.sleep(1)


def schedule_checks(checkers):
    schedule.clear()
    for checker in checkers:
        conf = checker.conf
        period = conf.get("period", 300)
        logger.info(
            "Scheduling checks for %r every %r seconds",
            conf["name"],
            period,
        )
        schedule.every(period).seconds.do(checker.check)


def execute_all(checkers):
    global interrupted
    for checker in checkers:
        if not interrupted:
            checker.check()


def on_reload_config(*args, **kwargs):
    global reload_conf_pending
    logger.info("Received SIGUSR1. Flagging configuration reload")
    reload_conf_pending = True


def on_backdoor(*args, **kwargs):
    global open_backdoor
    logger.info("Received SIGUSR2. Flagging backdoor to open")
    open_backdoor = True


def on_interrupt(*args, **kwargs):
    global interrupted
    interrupted = True



import sys
import logging
import time
from contextlib import contextmanager

from selenium import webdriver
from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
from xvfbwrapper import Xvfb
from ..conf import settings


logger = logging.getLogger(__name__)


firefox_instance = {
    'xvfb_display': None,
    'driver': None,
    'headed_driver': None,
}


def cleanup():
    """Must be called before exit"""
    global firefox_instance
    if firefox_instance['driver'] is not None:
        firefox_instance['driver'].quit()
        firefox_instance['driver'] = None
    if firefox_instance['headed_driver'] is not None:
        firefox_instance['headed_driver'].quit()
        firefox_instance['headed_driver'] = None
    if firefox_instance['xvfb_display'] is not None:
        firefox_instance['xvfb_display'].stop()
        firefox_instance['xvfb_display'] = None


def firefox_fetcher(conf):
    url = conf['url']
    delay = conf.get('delay')
    scenario = conf.get('scenario')
    headless = conf.get('headless', True)
    with firefox(headless) as driver:
        driver.get(url)
        if scenario:
            run_scenario(driver, scenario)
        if delay:
            time.sleep(delay)
        html = driver.find_element_by_xpath("//*").get_attribute("outerHTML")
        return True, html


@contextmanager
def firefox(headless=True):
    global firefox_instance
    if headless:
        if firefox_instance['xvfb_display'] is None:
            firefox_instance['xvfb_display'] = virtual_buffer()
        driver_key = 'driver'
    else:
        driver_key = 'headed_driver'
    if firefox_instance[driver_key] is None:
        if logger.level == logging.DEBUG:
            firefox_binary = FirefoxBinary(log_file=sys.stdout)
        else:
            firefox_binary = None
        firefox_instance[driver_key] = webdriver.Firefox(
            firefox_binary=firefox_binary,
        )
        firefox_instance[driver_key].set_window_size(1024, 768)
    yield firefox_instance[driver_key]


def virtual_buffer():
    """
    Try to start xvfb, trying multiple (up to 5) times if a failure
    """
    for _ in range(0, 6):
        xvfb_display = Xvfb()
        xvfb_display.start()
        # If Xvfb started, return.
        if xvfb_display.proc is not None:
            return xvfb_display
    raise Exception("Xvfb could not be started after six attempts.")


def run_scenario(driver, code):
    logger.info("Executing custom scenario")
    logger.debug(code)
    exec(code, {'driver': driver, 'creds': settings().creds})



import re


RE_VALID_FILENAME = re.compile(r'(?u)[^-\w.]')


def normalize_filename(text):
    """
    ORIGIN: https://github.com/django/django/blob/master/django/utils/text.py

    Returns the given string converted to a string that can be used for a clean
    filename. Specifically, leading and trailing spaces are removed; other
    spaces are converted to underscores; and anything that is not a unicode
    alphanumeric, dash, underscore, or dot, is removed.
    >>> get_valid_filename("john's portrait in 2004.jpg")
    'johns_portrait_in_2004.jpg'
    """
    return RE_VALID_FILENAME.sub('', text.strip().replace(' ', '_'))



import requests
from kibitzr.checker import Checker


def test_server_is_alive(target):
    """Sanity check, that test environment is properly setup"""
    response = requests.get("http://{0}:{1}/index.html".format(*target))
    assert response.status_code == 200


def test_simple_fetcher_with_pretty_json(target, json_conf):
    ok, content = Checker(json_conf).check()
    assert ok is True
    assert content == (
        '{\n'
        '  "first name": "Peter",\n'
        '  "last name": "Demin"\n'
        '}'
    )


def test_tag_transformer(target, html_text_conf):
    html_text_conf['transform'].insert(0, {
        'tag': 'div',
    })
    ok, content = Checker(html_text_conf).check()
    assert ok is True
    assert content == 'Hello world!'


def test_browser_css(target, html_text_conf):
    html_text_conf['transform'].insert(0, {
        'css': '.footer',
    })
    ok, content = Checker(html_text_conf).check()
    assert ok is True
    assert content == 'Footer content'


def test_browser_xpath(target, html_text_conf):
    html_text_conf['transform'].insert(0, {
        'xpath': './/*[@class="footer"]',
    })
    ok, content = Checker(html_text_conf).check()
    assert ok is True
    assert content == 'Footer content'


def test_scenario(target, html_text_conf):
    html_text_conf.update({
        'scenario': 'driver.find_element_by_id("page-link").click()'
    })
    ok, content = Checker(html_text_conf).check()
    assert ok is True
    assert content == 'Another page'


def test_valid_http_404(target, not_found_conf):
    not_found_conf.update({
        'valid_http': [404],
    })
    ok, content = Checker(not_found_conf).check()
    assert ok is True
    assert '404' in content



#!/usr/bin/env python

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

setup(
    name='kibitzr',
    version='2.7.1',
    description="Self hosted web page changes monitoring",
    long_description=readme + '\n\n' + history,
    author="Peter Demin",
    author_email='kibitzrrr@gmail.com',
    url='https://github.com/kibitzr/kibitzr',
    packages=[
        'kibitzr',
    ],
    package_dir={
        'kibitzr': 'kibitzr',
    },
    entry_points={
        'console_scripts': [
            'kibitzr=kibitzr.cli:entry'
        ]
    },
    include_package_data=True,
    install_requires=[
        'Click>=6.0',
        'requests',
        'schedule',
        'sh',
        'pyyaml',
        'selenium>=2.0, <3.0',
        'xvfbwrapper',
        'bs4',
        'six',
        'lxml',
        'lazy-object-proxy',
    ],
    license="MIT license",
    zip_safe=False,
    keywords='kibitzr',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest',
        'pytest-pep8',
        'pylint',
        'mock',
        'pytest-mock',
    ],
)



import json
import logging

import requests

from ..conf import settings


logger = logging.getLogger(__name__)


class SlackSession(object):
    def __init__(self, *args, **kwargs):
        slack = settings().notifiers.get('slack', {})
        slack.update(settings().creds.get('slack', {}))
        self.session = requests.Session()
        self.url = slack['url']

    def post(self, report, **kwargs):
        response = self.session.post(
            url=self.url,
            data={"payload": json.dumps({"text": report})},
        )
        logger.debug(response.text)
        response.raise_for_status()
        return response
